import { Injectable, Inject, NgModule } from '@angular/core';
import { HttpHeaders, HttpParams, HttpClient, HttpClientModule } from '@angular/common/http';
import { throwError, of } from 'rxjs';
import { map, catchError, mergeMap } from 'rxjs/operators';
import { parse } from 'url';
import { sha256 } from 'hash.js';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 */
class ResourceArray {
    constructor() {
        this.totalElements = 0;
        this.totalPages = 1;
        this.pageNumber = 1;
        this.result = [];
    }
    /**
     * @private
     * @param {?} query
     * @param {?} field
     * @param {?} value
     * @return {?}
     */
    static replaceOrAdd(query, field, value) {
        if (query) {
            /** @type {?} */
            const idx = query.indexOf(field);
            /** @type {?} */
            const idxNextAmp = query.indexOf('&', idx) === -1 ? query.indexOf('/', idx) : query.indexOf('&', idx);
            if (idx !== -1) {
                /** @type {?} */
                const searchValue = query.substring(idx, idxNextAmp);
                query = query.replace(searchValue, field + '=' + value);
            }
            else {
                query = query.concat('&' + field + '=' + value);
            }
        }
        else {
            query = '?' + field + '=' + value;
        }
        return query;
    }
    /**
     * @param {?} el
     * @return {?}
     */
    push(el) {
        this.result.push(el);
    }
    /**
     * @return {?}
     */
    length() {
        return this.result.length;
    }
    // Load next page
    /**
     * @param {?} type
     * @return {?}
     */
    next(type) {
        if (this.next_uri) {
            return ResourceHelper.getHttp().get(ResourceHelper.getProxy(this.next_uri), {
                headers: ResourceHelper.headers,
                observe: 'response'
            }).pipe(map((/**
             * @param {?} response
             * @return {?}
             */
            response => this.init(type, response, this.sortInfo))), catchError((/**
             * @param {?} error
             * @return {?}
             */
            error => throwError(error))));
        }
        return throwError('no next defined');
    }
    /**
     * @param {?} type
     * @return {?}
     */
    prev(type) {
        if (this.prev_uri) {
            return ResourceHelper.getHttp().get(ResourceHelper.getProxy(this.prev_uri), {
                headers: ResourceHelper.headers,
                observe: 'response'
            }).pipe(map((/**
             * @param {?} response
             * @return {?}
             */
            response => this.init(type, response, this.sortInfo))), catchError((/**
             * @param {?} error
             * @return {?}
             */
            error => throwError(error))));
        }
        return throwError('no prev defined');
    }
    /**
     * @param {?} type
     * @return {?}
     */
    first(type) {
        if (this.first_uri) {
            return ResourceHelper.getHttp().get(ResourceHelper.getProxy(this.first_uri), {
                headers: ResourceHelper.headers,
                observe: 'response'
            }).pipe(map((/**
             * @param {?} response
             * @return {?}
             */
            response => this.init(type, response, this.sortInfo))), catchError((/**
             * @param {?} error
             * @return {?}
             */
            error => throwError(error))));
        }
        return throwError('no first defined');
    }
    // Load first page
    /**
     * @param {?} type
     * @return {?}
     */
    last(type) {
        if (this.last_uri) {
            return ResourceHelper.getHttp().get(ResourceHelper.getProxy(this.last_uri), {
                headers: ResourceHelper.headers,
                observe: 'response'
            }).pipe(map((/**
             * @param {?} response
             * @return {?}
             */
            response => this.init(type, response, this.sortInfo))), catchError((/**
             * @param {?} error
             * @return {?}
             */
            error => throwError(error))));
        }
        return throwError('no last defined');
    }
    // Load last page
    /**
     * @param {?} type
     * @param {?} pageNumber
     * @return {?}
     */
    page(type, pageNumber) {
        this.self_uri = this.self_uri.replace('{?page,size,sort,projection}', '');
        this.self_uri = this.self_uri.replace('{&sort}', '');
        /** @type {?} */
        const urlParsed = parse(ResourceHelper.getProxy(this.self_uri));
        /** @type {?} */
        let query = ResourceArray.replaceOrAdd(urlParsed.query, 'size', this.pageSize.toString());
        query = ResourceArray.replaceOrAdd(query, 'page', pageNumber.toString());
        /** @type {?} */
        let uri = urlParsed.query ?
            ResourceHelper.getProxy(this.self_uri).replace(urlParsed.query, query) : ResourceHelper.getProxy(this.self_uri).concat(query);
        uri = this.addSortInfo(uri);
        return ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response'
        }).pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => this.init(type, response, this.sortInfo))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    // Load page with given pageNumber
    /**
     * @param {?} type
     * @param {...?} sort
     * @return {?}
     */
    sortElements(type, ...sort) {
        this.self_uri = this.self_uri.replace('{?page,size,sort}', '');
        this.self_uri = this.self_uri.replace('{&sort}', '');
        /** @type {?} */
        let uri = ResourceHelper.getProxy(this.self_uri).concat('?', 'size=', this.pageSize.toString(), '&page=', this.pageNumber.toString());
        uri = this.addSortInfo(uri);
        return ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response'
        }).pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => this.init(type, response, sort))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    // Sort collection based on given sort attribute
    /**
     * @param {?} type
     * @param {?} size
     * @return {?}
     */
    size(type, size) {
        /** @type {?} */
        let uri = ResourceHelper.getProxy(this.self_uri).concat('?', 'size=', size.toString());
        uri = this.addSortInfo(uri);
        return ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response'
        }).pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => this.init(type, response, this.sortInfo))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    // Load page with given size
    /**
     * @private
     * @param {?} type
     * @param {?} response
     * @param {?} sortInfo
     * @return {?}
     */
    init(type, response, sortInfo) {
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(this._embedded);
        result.sortInfo = sortInfo;
        ResourceHelper.instantiateResourceCollection(type, response, result);
        return result;
    }
    /**
     * @private
     * @param {?} uri
     * @return {?}
     */
    addSortInfo(uri) {
        if (this.sortInfo) {
            for (const item of this.sortInfo) {
                uri = uri.concat('&sort=', item.path, ',', item.order);
            }
        }
        return uri;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class Utils {
    /**
     * @param {?} value
     * @return {?}
     */
    static isNullOrUndefined(value) {
        return value === null || value === undefined;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    static isPrimitive(value) {
        return (typeof value !== 'object' && typeof value !== 'function') || value === null;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// @dynamic
class ResourceHelper {
    /**
     * @return {?}
     */
    static get headers() {
        if (Utils.isNullOrUndefined(this._headers)) {
            this._headers = new HttpHeaders();
        }
        return this._headers;
    }
    /**
     * @param {?} headers
     * @return {?}
     */
    static set headers(headers) {
        this._headers = headers;
    }
    /**
     * @param {?} params
     * @param {?=} options
     * @return {?}
     */
    static optionParams(params, options) {
        if (options) {
            params = this.params(params, options.params);
            if (options.size) {
                params = params.append('size', options.size.toString());
            }
            if (options.sort) {
                for (const s of options.sort) {
                    /** @type {?} */
                    let sortString = '';
                    sortString = s.path ? sortString.concat(s.path) : sortString;
                    sortString = s.order ? sortString.concat(',').concat(s.order) : sortString;
                    params = params.append('sort', sortString);
                }
            }
        }
        return params;
    }
    /**
     * @param {?} httpParams
     * @param {?=} params
     * @return {?}
     */
    static params(httpParams, params) {
        if (params) {
            for (const param of params) {
                httpParams = httpParams.append(param.key, param.value.toString());
            }
        }
        return httpParams;
    }
    /**
     * @param {?} resource
     * @return {?}
     */
    static resolveRelations(resource) {
        /** @type {?} */
        const result = {};
        for (const key in resource) {
            if (!Utils.isNullOrUndefined(resource[key])) {
                if (ResourceHelper.className(resource[key])
                    .find((/**
                 * @param {?} className
                 * @return {?}
                 */
                (className) => className === 'Resource')) || resource[key]._links) {
                    if (resource[key]._links) {
                        result[key] = resource[key]._links.self.href;
                    }
                }
                else if (Array.isArray(resource[key])) {
                    /** @type {?} */
                    const array = resource[key];
                    if (array) {
                        result[key] = [];
                        array.forEach((/**
                         * @param {?} element
                         * @return {?}
                         */
                        (element) => {
                            if (Utils.isPrimitive(element)) {
                                result[key].push(element);
                            }
                            else {
                                result[key].push(this.resolveRelations(element));
                            }
                        }));
                    }
                }
                else {
                    result[key] = resource[key];
                }
            }
        }
        return (/** @type {?} */ (result));
    }
    /**
     * @template T
     * @param {?} _embedded
     * @return {?}
     */
    static createEmptyResult(_embedded) {
        /** @type {?} */
        const resourceArray = new ResourceArray();
        resourceArray._embedded = _embedded;
        return resourceArray;
    }
    /**
     * @param {?} obj
     * @return {?}
     */
    static getClassName(obj) {
        /** @type {?} */
        const funcNameRegex = /function (.+?)\(/;
        /** @type {?} */
        const results = (funcNameRegex).exec(obj.constructor.toString());
        return (results && results.length > 1) ? results[1] : '';
    }
    /**
     * @param {?} objProto
     * @return {?}
     */
    static className(objProto) {
        /** @type {?} */
        const classNames = [];
        /** @type {?} */
        let obj = Object.getPrototypeOf(objProto);
        /** @type {?} */
        let className;
        // tslint:disable-next-line:no-conditional-assignment
        while ((className = ResourceHelper.getClassName(obj)) !== 'Object') {
            classNames.push(className);
            obj = Object.getPrototypeOf(obj);
        }
        return classNames;
    }
    /**
     * @template T
     * @param {?} entity
     * @param {?} response
     * @return {?}
     */
    static instantiateResourceFromResponse(entity, response) {
        if (response.status >= 200 && response.status <= 207) {
            return ResourceHelper.instantiateResource(entity, response.body);
        }
        else if (response.status === 404) {
            return null;
        }
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} response
     * @param {?} result
     * @param {?=} builder
     * @return {?}
     */
    static instantiateResourceCollection(type, response, result, builder) {
        if (response.status >= 200 && response.status <= 207) {
            /** @type {?} */
            const payload = response.body;
            if (payload[result._embedded]) {
                for (const embeddedClassName of Object.keys(payload[result._embedded])) {
                    /** @type {?} */
                    const embedded = payload[result._embedded];
                    /** @type {?} */
                    const items = embedded[embeddedClassName];
                    for (const item of items) {
                        /** @type {?} */
                        let instance = new type();
                        instance = this.searchSubtypes(builder, embeddedClassName, instance);
                        this.instantiateResource(instance, item);
                        result.push(instance);
                    }
                }
            }
            result.totalElements = payload.page ? payload.page.totalElements : result.length;
            result.totalPages = payload.page ? payload.page.totalPages : 1;
            result.pageNumber = payload.page ? payload.page.number : 1;
            result.pageSize = payload.page ? payload.page.size : 20;
            result.self_uri = payload._links && payload._links.self ? payload._links.self.href : undefined;
            result.next_uri = payload._links && payload._links.next ? payload._links.next.href : undefined; // tslint:disable-line:max-line-length
            result.prev_uri = payload._links && payload._links.prev ? payload._links.prev.href : undefined; // tslint:disable-line:max-line-length
            result.first_uri = payload._links && payload._links.first ? payload._links.first.href : undefined; // tslint:disable-line:max-line-length
            result.last_uri = payload._links && payload._links.last ? payload._links.last.href : undefined; // tslint:disable-line:max-line-length
        }
        else if (response.status === 404) {
            result.result = [];
        }
        return result;
    }
    /**
     * @template T
     * @param {?} builder
     * @param {?} embeddedClassName
     * @param {?} instance
     * @return {?}
     */
    static searchSubtypes(builder, embeddedClassName, instance) {
        if (builder && builder.subtypes) {
            /** @type {?} */
            const keys = builder.subtypes.keys();
            Array.from(keys).forEach((/**
             * @param {?} subtypeKey
             * @return {?}
             */
            (subtypeKey) => {
                if (embeddedClassName.toLowerCase().startsWith(subtypeKey.toLowerCase())) {
                    /** @type {?} */
                    const subtype = builder.subtypes.get(subtypeKey);
                    instance = new subtype();
                }
            }));
        }
        return instance;
    }
    /**
     * @template T
     * @param {?} entity
     * @param {?} payload
     * @return {?}
     */
    static instantiateResource(entity, payload) {
        for (const p of Object.keys(payload)) {
            // TODO array initClearCacheProcess
            /* if(entity[p].constructor === Array && isNullOrUndefined(payload[p]))
                 entity[p] = [];
             else*/
            entity[p] = payload[p];
        }
        return entity;
    }
    /**
     * @param {?} proxy_uri
     * @return {?}
     */
    static setProxyUri(proxy_uri) {
        ResourceHelper.proxy_uri = proxy_uri;
    }
    /**
     * @param {?} root_uri
     * @return {?}
     */
    static setRootUri(root_uri) {
        ResourceHelper.root_uri = root_uri;
    }
    /**
     * @return {?}
     */
    static getURL() {
        return ResourceHelper.proxy_uri && ResourceHelper.proxy_uri !== '' ?
            ResourceHelper.addSlash(ResourceHelper.proxy_uri) :
            ResourceHelper.addSlash(ResourceHelper.root_uri);
    }
    /**
     * @param {?} uri
     * @return {?}
     */
    static getProxy(uri) {
        uri = uri.replace('{?projection}', '');
        if (!ResourceHelper.proxy_uri || ResourceHelper.proxy_uri === '') {
            return uri;
        }
        return ResourceHelper.addSlash(uri.replace(ResourceHelper.root_uri, ResourceHelper.proxy_uri));
    }
    /**
     * @param {?} http
     * @return {?}
     */
    static setHttp(http) {
        this.http = http;
    }
    /**
     * @return {?}
     */
    static getHttp() {
        return this.http;
    }
    /**
     * @return {?}
     */
    static getRootUri() {
        return this.root_uri;
    }
    /**
     * @private
     * @param {?} uri
     * @return {?}
     */
    static addSlash(uri) {
        /** @type {?} */
        const uriParsed = parse(uri);
        if (Utils.isNullOrUndefined(uriParsed.search) && uri && uri[uri.length - 1] !== '/') {
            return uri + '/';
        }
        return uri;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {number} */
const EvictStrategy = {
    EvictTrivial: 0,
    EvictSmart: 1,
};
EvictStrategy[EvictStrategy.EvictTrivial] = 'EvictTrivial';
EvictStrategy[EvictStrategy.EvictSmart] = 'EvictSmart';
// @dynamic
class CacheHelper {
    /**
     * @return {?}
     */
    static initClearCacheProcess() {
        if (this.isActive) {
            setInterval((/**
             * @return {?}
             */
            () => {
                if (CacheHelper.evictStrategy === EvictStrategy.EvictTrivial) {
                    this.evictAll();
                }
                else if (CacheHelper.evictStrategy === EvictStrategy.EvictSmart) {
                    this.cacheMap.forEach((/**
                     * @param {?} value
                     * @param {?} key
                     * @return {?}
                     */
                    (value, key) => {
                        if (value.expire > 0 && Date.now() > value.expire) {
                            this.evict(key);
                        }
                    }));
                }
            }), 15 * 60 * 1000);
        }
    }
    /**
     * @template T
     * @param {?} link
     * @param {?=} body
     * @param {?=} params
     * @param {?=} isActiveLocal
     * @return {?}
     */
    static ifPresent(link, body, params, isActiveLocal = true) {
        if (!this.isActive || !isActiveLocal) {
            return false;
        }
        return this.cacheMap.has(CacheHelper.key(link, body, params));
    }
    /**
     * @template T
     * @param {?} link
     * @param {?=} body
     * @param {?=} params
     * @return {?}
     */
    static getArray(link, body, params) {
        return this.cacheMap.get(CacheHelper.key(link, body, params)).entity;
    }
    /**
     * @template T
     * @param {?} link
     * @param {?} array
     * @param {?=} expireMs
     * @param {?=} body
     * @param {?=} params
     * @return {?}
     */
    static putArray(link, array, expireMs = 10 * 60 * 1000, body, params) {
        if (this.isActive) {
            /** @type {?} */
            const resourceExpire = {
                entity: array,
                expire: CacheHelper.expireDate(expireMs)
            };
            this.cacheMap.set(CacheHelper.key(link, body, params), resourceExpire);
        }
    }
    /**
     * @template T
     * @param {?} link
     * @param {?=} body
     * @param {?=} params
     * @return {?}
     */
    static get(link, body, params) {
        return this.cacheMap.get(CacheHelper.key(link, body, params)).entity;
    }
    /**
     * @template T
     * @param {?} link
     * @param {?} array
     * @param {?=} expireMs
     * @param {?=} body
     * @param {?=} params
     * @return {?}
     */
    static put(link, array, expireMs = 10 * 60 * 1000, body, params) {
        if (this.isActive) {
            /** @type {?} */
            const resourceExpire = {
                entity: array,
                expire: CacheHelper.expireDate(expireMs)
            };
            this.cacheMap.set(CacheHelper.key(link, body, params), resourceExpire);
        }
    }
    /**
     * @param {?} key
     * @return {?}
     */
    static evict(key) {
        this.cacheMap.delete(key);
    }
    /**
     * @return {?}
     */
    static evictAll() {
        this.cacheMap.clear();
    }
    /**
     * @private
     * @param {?} expireMs
     * @return {?}
     */
    static expireDate(expireMs) {
        if (expireMs === 0) {
            return 0;
        }
        return Date.now() + expireMs;
    }
    /**
     * @private
     * @param {?} link
     * @param {?=} body
     * @param {?=} halOptions
     * @return {?}
     */
    static key(link, body, halOptions) {
        /** @type {?} */
        let k = link;
        if (body) {
            k += body;
        }
        if (halOptions) {
            k += CacheHelper.toStringParams(halOptions);
        }
        /** @type {?} */
        const key = sha256().update(k).digest('hex');
        return key;
    }
    /**
     * @private
     * @param {?} options
     * @return {?}
     */
    static toStringParams(options) {
        /** @type {?} */
        let s = '';
        if (options.size) {
            s = 'size=' + options.size.toString() + '&';
        }
        if (options.notPaged) {
            s += 'notPaged=true&';
        }
        if (options.params) {
            options.params.forEach((/**
             * @param {?} param
             * @return {?}
             */
            param => {
                s += param.key + '=' + param.value + '&';
            }));
        }
        if (options.sort) {
            options.sort.forEach((/**
             * @param {?} sortInfo
             * @return {?}
             */
            sortInfo => {
                /** @type {?} */
                let sortString = '';
                sortString = sortInfo.path ? sortString.concat(sortInfo.path) : sortString;
                sortString = sortInfo.order ? sortString.concat(',').concat(sortInfo.order) : sortString;
                s += 'sort' + sortString + '&';
            }));
        }
        return s;
    }
}
CacheHelper.isActive = true;
// TODO
CacheHelper.maxEntries = 100;
CacheHelper.evictStrategy = EvictStrategy.EvictTrivial;
CacheHelper.defaultExpire = 10 * 60 * 1000; // 10 minutes
// 10 minutes
CacheHelper.cacheMap = new Map();

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class ExternalService {
    /**
     * @param {?} externalConfigurationService
     */
    constructor(externalConfigurationService) {
        this.externalConfigurationService = externalConfigurationService;
        ResourceHelper.setProxyUri(externalConfigurationService.getProxyUri());
        ResourceHelper.setRootUri(externalConfigurationService.getRootUri());
        ResourceHelper.setHttp(externalConfigurationService.getHttp());
        CacheHelper.initClearCacheProcess();
    }
    /**
     * @param {?} externalConfigurationService
     * @return {?}
     */
    updateExternalConfigurationHandlerInterface(externalConfigurationService) {
        this.externalConfigurationService = externalConfigurationService;
        ResourceHelper.setProxyUri(externalConfigurationService.getProxyUri());
        ResourceHelper.setRootUri(externalConfigurationService.getRootUri());
        ResourceHelper.setHttp(externalConfigurationService.getHttp());
    }
    /**
     * @return {?}
     */
    getExternalConfiguration() {
        return this.externalConfigurationService.getExternalConfiguration();
    }
    /**
     * @return {?}
     */
    getProxyUri() {
        return this.externalConfigurationService.getProxyUri();
    }
    /**
     * @return {?}
     */
    getRootUri() {
        return this.externalConfigurationService.getRootUri();
    }
    /**
     * @return {?}
     */
    getURL() {
        return ResourceHelper.getURL();
    }
    /**
     * @return {?}
     */
    getHttp() {
        return ResourceHelper.getHttp();
    }
}
ExternalService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
ExternalService.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: ['ExternalConfigurationService',] }] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class CustomEncoder {
    /**
     * @param {?} key
     * @return {?}
     */
    encodeKey(key) {
        return encodeURIComponent(key);
    }
    /**
     * @param {?} value
     * @return {?}
     */
    encodeValue(value) {
        return encodeURIComponent(value);
    }
    /**
     * @param {?} key
     * @return {?}
     */
    decodeKey(key) {
        return decodeURIComponent(key);
    }
    /**
     * @param {?} value
     * @return {?}
     */
    decodeValue(value) {
        return decodeURIComponent(value);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class ResourceService {
    /**
     * @param {?} externalService
     */
    constructor(externalService) {
        this.externalService = externalService;
    }
    /**
     * @private
     * @return {?}
     */
    static getURL() {
        return ResourceHelper.getURL();
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resource
     * @param {?} _embedded
     * @param {?=} options
     * @param {?=} subType
     * @return {?}
     */
    getAll(type, resource, _embedded, options, subType) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource);
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams({ encoder: new CustomEncoder() }), options);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(_embedded);
        this.setUrls(result);
        result.sortInfo = options ? options.sort : undefined;
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response',
            params
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result, subType))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resource
     * @param {?} id
     * @param {?=} params
     * @param {?=} builder
     * @return {?}
     */
    get(type, resource, id, params, builder) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('/', id);
        /** @type {?} */
        let result = new type();
        /** @type {?} */
        const httpParams = ResourceHelper.params(new HttpParams(), params);
        this.setUrlsResource(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response',
            params: httpParams
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        (response) => {
            if (builder) {
                /** @type {?} */
                const linkHref = parse(response.body._links.self.href).pathname;
                /** @type {?} */
                const regex = /([A-Za-z0-9]+)\/([A-Za-z0-9]+)\/([A-Za-z0-9]+)/g;
                /** @type {?} */
                const match = regex.exec(linkHref);
                if (match != null) {
                    /** @type {?} */
                    const embeddedClassName = match[2];
                    result = ResourceHelper.searchSubtypes(builder, embeddedClassName, result);
                }
            }
            return ResourceHelper.instantiateResourceFromResponse(result, response);
        })), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resourceLink
     * @return {?}
     */
    getBySelfLink(type, resourceLink) {
        /** @type {?} */
        const result = new type();
        this.setUrlsResource(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(ResourceHelper.getProxy(resourceLink), { headers: ResourceHelper.headers });
        return observable.pipe(map((/**
         * @param {?} data
         * @return {?}
         */
        data => ResourceHelper.instantiateResource(result, data))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} query
     * @param {?} resource
     * @param {?} _embedded
     * @param {?=} options
     * @param {?=} subType
     * @return {?}
     */
    search(type, query, resource, _embedded, options, subType) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('/search/', query);
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams({ encoder: new CustomEncoder() }), options);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(_embedded);
        this.setUrls(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response',
            params
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result, subType))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} query
     * @param {?} resource
     * @param {?=} options
     * @return {?}
     */
    searchSingle(type, query, resource, options) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('/search/', query);
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams({ encoder: new CustomEncoder() }), options);
        /** @type {?} */
        const result = new type();
        this.setUrlsResource(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, { headers: ResourceHelper.headers, params });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResource(result, response))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} query
     * @param {?} resource
     * @param {?} _embedded
     * @param {?=} options
     * @param {?=} subType
     * @return {?}
     */
    customQuery(type, query, resource, _embedded, options, subType) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource + query);
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams({ encoder: new CustomEncoder() }), options);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(_embedded);
        this.setUrls(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response',
            params
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result, subType))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} query
     * @param {?} resource
     * @param {?} _embedded
     * @param {?=} options
     * @param {?=} body
     * @param {?=} subType
     * @return {?}
     */
    customQueryPost(type, query, resource, _embedded, options, body, subType) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource + query);
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams(), options);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(_embedded);
        this.setUrls(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().post(uri, body, {
            headers: ResourceHelper.headers,
            observe: 'response',
            params
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result, subType))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resourceLink
     * @return {?}
     */
    getByRelation(type, resourceLink) {
        /** @type {?} */
        const result = new type();
        this.setUrlsResource(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(resourceLink, { headers: ResourceHelper.headers });
        return observable.pipe(map((/**
         * @param {?} data
         * @return {?}
         */
        data => ResourceHelper.instantiateResource(result, data))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resourceLink
     * @param {?} _embedded
     * @param {?=} builder
     * @return {?}
     */
    getByRelationArray(type, resourceLink, _embedded, builder) {
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(_embedded);
        this.setUrls(result);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(resourceLink, {
            headers: ResourceHelper.headers,
            observe: 'response'
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result, builder))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resource
     * @param {?} id
     * @param {?} projectionName
     * @return {?}
     */
    getProjection(type, resource, id, projectionName) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('/', id).concat('?projection=' + projectionName);
        /** @type {?} */
        const result = new type();
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, { headers: ResourceHelper.headers });
        return observable.pipe(map((/**
         * @param {?} data
         * @return {?}
         */
        data => ResourceHelper.instantiateResource(result, data))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resource
     * @param {?} projectionName
     * @return {?}
     */
    getProjectionArray(type, resource, projectionName) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('?projection=' + projectionName);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult('_embedded');
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response'
        });
        return observable
            .pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error)))).pipe(map((/**
         * @param {?} resourceArray
         * @return {?}
         */
        (resourceArray) => {
            return resourceArray.result;
        })));
    }
    /**
     * @param {?} resource
     * @param {?=} query
     * @param {?=} options
     * @return {?}
     */
    count(resource, query, options) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('/search/' + (query === undefined ? 'countAll' : query));
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams(), options);
        return ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response',
            params
        }).pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        (response) => Number(response.body))), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} selfResource
     * @param {?} entity
     * @return {?}
     */
    create(selfResource, entity) {
        /** @type {?} */
        const uri = ResourceHelper.getURL() + selfResource;
        /** @type {?} */
        const payload = ResourceHelper.resolveRelations(entity);
        this.setUrlsResource(entity);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().post(uri, payload, {
            headers: ResourceHelper.headers,
            observe: 'response'
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        (response) => {
            if (response.status >= 200 && response.status <= 207) {
                return ResourceHelper.instantiateResource(entity, response.body);
            }
            else if (response.status === 500) {
                /** @type {?} */
                const body = response.body;
                return throwError(body.error);
            }
        })), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} entity
     * @return {?}
     */
    update(entity) {
        /** @type {?} */
        const uri = ResourceHelper.getProxy(entity._links.self.href);
        /** @type {?} */
        const payload = ResourceHelper.resolveRelations(entity);
        this.setUrlsResource(entity);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().put(uri, payload, {
            headers: ResourceHelper.headers,
            observe: 'response'
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        (response) => {
            if (response.status >= 200 && response.status <= 207) {
                return ResourceHelper.instantiateResource(entity, response.body);
            }
            else if (response.status === 500) {
                /** @type {?} */
                const body = response.body;
                return throwError(body.error);
            }
        })), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} entity
     * @return {?}
     */
    patch(entity) {
        /** @type {?} */
        const uri = ResourceHelper.getProxy(entity._links.self.href);
        /** @type {?} */
        const payload = ResourceHelper.resolveRelations(entity);
        this.setUrlsResource(entity);
        /** @type {?} */
        const observable = ResourceHelper.getHttp().patch(uri, payload, {
            headers: ResourceHelper.headers,
            observe: 'response'
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        (response) => {
            if (response.status >= 200 && response.status <= 207) {
                return ResourceHelper.instantiateResource(entity, response.body);
            }
            else if (response.status === 500) {
                /** @type {?} */
                const body = response.body;
                return throwError(body.error);
            }
        })), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} entity
     * @return {?}
     */
    delete(entity) {
        /** @type {?} */
        const uri = ResourceHelper.getProxy(entity._links.self.href);
        return ResourceHelper.getHttp().delete(uri, { headers: ResourceHelper.headers }).pipe(catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @return {?}
     */
    hasNext(resourceArray) {
        return resourceArray.next_uri !== undefined;
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @return {?}
     */
    hasPrev(resourceArray) {
        return resourceArray.prev_uri !== undefined;
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @return {?}
     */
    hasFirst(resourceArray) {
        return resourceArray.first_uri !== undefined;
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @return {?}
     */
    hasLast(resourceArray) {
        return resourceArray.last_uri !== undefined;
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @return {?}
     */
    next(resourceArray, type) {
        return resourceArray.next(type);
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @return {?}
     */
    prev(resourceArray, type) {
        return resourceArray.prev(type);
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @return {?}
     */
    first(resourceArray, type) {
        return resourceArray.first(type);
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @return {?}
     */
    last(resourceArray, type) {
        return resourceArray.last(type);
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @param {?} id
     * @return {?}
     */
    page(resourceArray, type, id) {
        return resourceArray.page(type, id);
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @param {...?} sort
     * @return {?}
     */
    sortElements(resourceArray, type, ...sort) {
        return resourceArray.sortElements(type, ...sort);
    }
    /**
     * @template T
     * @param {?} resourceArray
     * @param {?} type
     * @param {?} size
     * @return {?}
     */
    size(resourceArray, type, size) {
        return resourceArray.size(type, size);
    }
    /**
     * @private
     * @param {?=} resource
     * @return {?}
     */
    getResourceUrl(resource) {
        /** @type {?} */
        let resourceUrl = ResourceService.getURL();
        if (!resourceUrl.endsWith('/')) {
            resourceUrl = resourceUrl.concat('/');
        }
        if (resource) {
            return resourceUrl.concat(resource);
        }
        resourceUrl = resourceUrl.replace('{?projection}', '');
        return resourceUrl;
    }
    /**
     * @private
     * @template T
     * @param {?} result
     * @return {?}
     */
    setUrls(result) {
        result.proxyUrl = this.externalService.getProxyUri();
        result.rootUrl = this.externalService.getRootUri();
    }
    /**
     * @private
     * @template T
     * @param {?} result
     * @return {?}
     */
    setUrlsResource(result) {
        result.proxyUrl = this.externalService.getProxyUri();
        result.rootUrl = this.externalService.getRootUri();
    }
}
ResourceService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
ResourceService.ctorParameters = () => [
    { type: ExternalService }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 */
class RestService {
    /**
     * @param {?} type
     * @param {?} resource
     * @param {?} injector
     * @param {?=} _embedded
     */
    constructor(type, resource, injector, _embedded) {
        this.injector = injector;
        this._embedded = '_embedded';
        this.type = type;
        this.resource = resource;
        this.resourceService = injector.get(ResourceService);
        if (!Utils.isNullOrUndefined(_embedded)) {
            this._embedded = _embedded;
        }
    }
    /**
     * @protected
     * @param {?} error
     * @return {?}
     */
    static handleError(error) {
        return throwError(error);
    }
    /**
     * @param {?=} options
     * @param {?=} subType
     * @return {?}
     */
    getAll(options, subType) {
        return this.resourceService.getAll(this.type, this.resource, this._embedded, options, subType).pipe(mergeMap((/**
         * @param {?} resourceArray
         * @return {?}
         */
        (resourceArray) => {
            if (options && options.notPaged && !Utils.isNullOrUndefined(resourceArray.first_uri)) {
                options.notPaged = false;
                options.size = resourceArray.totalElements;
                return this.getAll(options);
            }
            else {
                this.resourceArray = resourceArray;
                return of(resourceArray.result);
            }
        })));
    }
    /**
     * @param {?} id
     * @param {?=} params
     * @param {?=} builder
     * @return {?}
     */
    get(id, params, builder) {
        return this.resourceService.get(this.type, this.resource, id, params, builder);
    }
    /**
     * @param {?} selfLink
     * @return {?}
     */
    getBySelfLink(selfLink) {
        return this.resourceService.getBySelfLink(this.type, selfLink);
    }
    /**
     * @param {?} query
     * @param {?=} options
     * @param {?=} subType
     * @return {?}
     */
    search(query, options, subType) {
        return this.resourceService.search(this.type, query, this.resource, this._embedded, options, subType).pipe(mergeMap((/**
         * @param {?} resourceArray
         * @return {?}
         */
        (resourceArray) => {
            if (options && options.notPaged && !Utils.isNullOrUndefined(resourceArray.first_uri)) {
                options.notPaged = false;
                options.size = resourceArray.totalElements;
                return this.search(query, options, subType);
            }
            else {
                this.resourceArray = resourceArray;
                return of(resourceArray.result);
            }
        })));
    }
    /**
     * @param {?} query
     * @param {?=} options
     * @return {?}
     */
    searchSingle(query, options) {
        return this.resourceService.searchSingle(this.type, query, this.resource, options);
    }
    /**
     * @param {?} query
     * @param {?=} options
     * @param {?=} subType
     * @return {?}
     */
    customQuery(query, options, subType) {
        return this.resourceService.customQuery(this.type, query, this.resource, this._embedded, options, subType).pipe(mergeMap((/**
         * @param {?} resourceArray
         * @return {?}
         */
        (resourceArray) => {
            if (options && options.notPaged && !Utils.isNullOrUndefined(resourceArray.first_uri)) {
                options.notPaged = false;
                options.size = resourceArray.totalElements;
                return this.customQuery(query, options, subType);
            }
            else {
                this.resourceArray = resourceArray;
                return of(resourceArray.result);
            }
        })));
    }
    /**
     * @param {?} query
     * @param {?=} options
     * @param {?=} body
     * @param {?=} subType
     * @return {?}
     */
    customQueryPost(query, options, body, subType) {
        return this.resourceService.customQueryPost(this.type, query, this.resource, this._embedded, options, body, subType).pipe(mergeMap((/**
         * @param {?} resourceArray
         * @return {?}
         */
        (resourceArray) => {
            if (options && options.notPaged && !Utils.isNullOrUndefined(resourceArray.first_uri)) {
                options.notPaged = false;
                options.size = resourceArray.totalElements;
                return this.customQueryPost(query, options, body, subType);
            }
            else {
                this.resourceArray = resourceArray;
                return of(resourceArray.result);
            }
        })));
    }
    /**
     * @param {?} relation
     * @param {?=} builder
     * @return {?}
     */
    getByRelationArray(relation, builder) {
        return this.resourceService.getByRelationArray(this.type, relation, this._embedded, builder).pipe(map((/**
         * @param {?} resourceArray
         * @return {?}
         */
        (resourceArray) => {
            this.resourceArray = resourceArray;
            return resourceArray.result;
        })));
    }
    /**
     * @param {?} relation
     * @return {?}
     */
    getByRelation(relation) {
        return this.resourceService.getByRelation(this.type, relation);
    }
    /**
     * @param {?=} query
     * @param {?=} options
     * @return {?}
     */
    count(query, options) {
        return this.resourceService.count(this.resource, query, options);
    }
    /**
     * @param {?} entity
     * @return {?}
     */
    create(entity) {
        return this.resourceService.create(this.resource, entity);
    }
    /**
     * @param {?} entity
     * @return {?}
     */
    update(entity) {
        return this.resourceService.update(entity);
    }
    /**
     * @param {?} entity
     * @return {?}
     */
    patch(entity) {
        return this.resourceService.patch(entity);
    }
    /**
     * @param {?} entity
     * @return {?}
     */
    delete(entity) {
        return this.resourceService.delete(entity);
    }
    /**
     * @return {?}
     */
    totalElement() {
        if (this.resourceArray && this.resourceArray.totalElements) {
            return this.resourceArray.totalElements;
        }
        return 0;
    }
    /**
     * @return {?}
     */
    totalPages() {
        if (this.resourceArray && this.resourceArray.totalPages) {
            return this.resourceArray.totalPages;
        }
        return 1;
    }
    /**
     * @return {?}
     */
    hasFirst() {
        if (this.resourceArray) {
            return this.resourceService.hasFirst(this.resourceArray);
        }
        return false;
    }
    /**
     * @return {?}
     */
    hasNext() {
        if (this.resourceArray) {
            return this.resourceService.hasNext(this.resourceArray);
        }
        return false;
    }
    /**
     * @return {?}
     */
    hasPrev() {
        if (this.resourceArray) {
            return this.resourceService.hasPrev(this.resourceArray);
        }
        return false;
    }
    /**
     * @return {?}
     */
    hasLast() {
        if (this.resourceArray) {
            return this.resourceService.hasLast(this.resourceArray);
        }
        return false;
    }
    /**
     * @return {?}
     */
    next() {
        if (this.resourceArray) {
            return this.resourceService.next(this.resourceArray, this.type).pipe(map((/**
             * @param {?} resourceArray
             * @return {?}
             */
            (resourceArray) => {
                this.resourceArray = resourceArray;
                return resourceArray.result;
            })));
        }
        else {
            return throwError('no resourceArray found');
        }
    }
    /**
     * @return {?}
     */
    prev() {
        if (this.resourceArray) {
            return this.resourceService.prev(this.resourceArray, this.type).pipe(map((/**
             * @param {?} resourceArray
             * @return {?}
             */
            (resourceArray) => {
                this.resourceArray = resourceArray;
                return resourceArray.result;
            })));
        }
        else {
            return throwError('no resourceArray found');
        }
    }
    /**
     * @return {?}
     */
    first() {
        if (this.resourceArray) {
            return this.resourceService.first(this.resourceArray, this.type)
                .pipe(map((/**
             * @param {?} resourceArray
             * @return {?}
             */
            (resourceArray) => {
                this.resourceArray = resourceArray;
                return resourceArray.result;
            })));
        }
        else {
            return throwError('no resourceArray found');
        }
    }
    /**
     * @return {?}
     */
    last() {
        if (this.resourceArray) {
            return this.resourceService.last(this.resourceArray, this.type)
                .pipe(map((/**
             * @param {?} resourceArray
             * @return {?}
             */
            (resourceArray) => {
                this.resourceArray = resourceArray;
                return resourceArray.result;
            })));
        }
        else {
            return throwError('no resourceArray found');
        }
    }
    /**
     * @param {?} pageNumber
     * @return {?}
     */
    page(pageNumber) {
        if (this.resourceArray) {
            return this.resourceService.page(this.resourceArray, this.type, pageNumber).pipe(map((/**
             * @param {?} resourceArray
             * @return {?}
             */
            (resourceArray) => {
                this.resourceArray = resourceArray;
                return resourceArray.result;
            })));
        }
        else {
            return throwError('no resourceArray found');
        }
    }
    /**
     * @protected
     * @param {?} error
     * @return {?}
     */
    handleError(error) {
        return RestService.handleError(error);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 */
class Resource {
    constructor() {
    }
    /**
     * @return {?}
     */
    get subtypes() {
        return this._subtypes;
    }
    /**
     * @param {?} _subtypes
     * @return {?}
     */
    set subtypes(_subtypes) {
        this._subtypes = _subtypes;
    }
    // Get related resource
    /**
     * @template T
     * @param {?} type
     * @param {?} relation
     * @param {?=} builder
     * @param {?=} expireMs
     * @param {?=} isCacheActive
     * @return {?}
     */
    getRelation(type, relation, builder, expireMs = CacheHelper.defaultExpire, isCacheActive = true) {
        /** @type {?} */
        let result = new type();
        if (this.existRelationLink(relation)) {
            if (CacheHelper.ifPresent(this.getRelationLinkHref(relation), null, null, isCacheActive)) {
                return of(CacheHelper.get(this.getRelationLinkHref(relation)));
            }
            /** @type {?} */
            const observable = ResourceHelper.getHttp().get(ResourceHelper.getProxy(this.getRelationLinkHref(relation)), { headers: ResourceHelper.headers });
            return observable.pipe(map((/**
             * @param {?} data
             * @return {?}
             */
            (data) => {
                if (builder) {
                    for (const embeddedClassName of Object.keys(data._links)) {
                        if (embeddedClassName === 'self') {
                            /** @type {?} */
                            const href = data._links[embeddedClassName].href;
                            /** @type {?} */
                            const idx = href.lastIndexOf('/');
                            /** @type {?} */
                            const realClassName = href.replace(ResourceHelper.getRootUri(), '').substring(0, idx);
                            result = ResourceHelper.searchSubtypes(builder, realClassName, result);
                            break;
                        }
                    }
                }
                /** @type {?} */
                const resource = ResourceHelper.instantiateResource(result, data);
                CacheHelper.put(this.getRelationLinkHref(relation), resource, expireMs);
                return resource;
            })));
        }
        else {
            return of(null);
        }
    }
    // Get collection of related resources
    /**
     * @template T
     * @param {?} type
     * @param {?} relation
     * @param {?=} _embedded
     * @param {?=} options
     * @param {?=} builder
     * @param {?=} expireMs
     * @param {?=} isCacheActive
     * @return {?}
     */
    getRelationArray(type, relation, _embedded, options, builder, expireMs = CacheHelper.defaultExpire, isCacheActive = true) {
        /** @type {?} */
        const params = ResourceHelper.optionParams(new HttpParams({ encoder: new CustomEncoder() }), options);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult(Utils.isNullOrUndefined(_embedded) ? '_embedded' : _embedded);
        if (this.existRelationLink(relation)) {
            if (CacheHelper.ifPresent(this.getRelationLinkHref(relation), null, options, isCacheActive)) {
                return of(CacheHelper.getArray(this.getRelationLinkHref(relation)));
            }
            /** @type {?} */
            const observable = ResourceHelper.getHttp().get(ResourceHelper.getProxy(this.getRelationLinkHref(relation)), {
                headers: ResourceHelper.headers, observe: 'response', params
            });
            return observable
                .pipe(map((/**
             * @param {?} response
             * @return {?}
             */
            response => ResourceHelper.instantiateResourceCollection(type, response, result, builder))), catchError((/**
             * @param {?} error
             * @return {?}
             */
            error => throwError(error)))).pipe(map((/**
             * @param {?} array
             * @return {?}
             */
            (array) => {
                CacheHelper.putArray(this.getRelationLinkHref(relation), array.result, expireMs);
                return array.result;
            })));
        }
        else {
            return of([]);
        }
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resource
     * @param {?} id
     * @param {?} projectionName
     * @param {?=} expireMs
     * @param {?=} isCacheActive
     * @return {?}
     */
    getProjection(type, resource, id, projectionName, expireMs = CacheHelper.defaultExpire, isCacheActive = true) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('/', id).concat('?projection=' + projectionName);
        /** @type {?} */
        const result = new type();
        if (CacheHelper.ifPresent(uri, null, null, isCacheActive)) {
            return of(CacheHelper.get(uri));
        }
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, { headers: ResourceHelper.headers });
        return observable.pipe(map((/**
         * @param {?} data
         * @return {?}
         */
        data => {
            /** @type {?} */
            const resourceObject = ResourceHelper.instantiateResource(result, data);
            CacheHelper.put(uri, resourceObject, expireMs);
            return resourceObject;
        })), catchError((/**
         * @param {?} error
         * @return {?}
         */
        error => throwError(error))));
    }
    /**
     * @template T
     * @param {?} type
     * @param {?} resource
     * @param {?} projectionName
     * @param {?=} expireMs
     * @param {?=} isCacheActive
     * @return {?}
     */
    getProjectionArray(type, resource, projectionName, expireMs = CacheHelper.defaultExpire, isCacheActive = true) {
        /** @type {?} */
        const uri = this.getResourceUrl(resource).concat('?projection=' + projectionName);
        /** @type {?} */
        const result = ResourceHelper.createEmptyResult('_embedded');
        if (CacheHelper.ifPresent(uri, null, null, isCacheActive)) {
            return of(CacheHelper.getArray(uri));
        }
        /** @type {?} */
        const observable = ResourceHelper.getHttp().get(uri, {
            headers: ResourceHelper.headers,
            observe: 'response'
        });
        return observable.pipe(map((/**
         * @param {?} response
         * @return {?}
         */
        response => ResourceHelper.instantiateResourceCollection(type, response, result))), map((/**
         * @param {?} array
         * @return {?}
         */
        (array) => {
            CacheHelper.putArray(uri, array.result, expireMs);
            return array.result;
        })));
    }
    // Adds the given resource to the bound collection by the relation
    /**
     * @template T
     * @param {?} relation
     * @param {?} resource
     * @return {?}
     */
    addRelation(relation, resource) {
        if (this.existRelationLink(relation)) {
            /** @type {?} */
            const header = ResourceHelper.headers.append('Content-Type', 'text/uri-list');
            return ResourceHelper.getHttp().put(ResourceHelper.getProxy(this.getRelationLinkHref(relation)), resource._links.self.href, { headers: header });
        }
        else {
            return throwError('no relation found');
        }
    }
    // Bind the given resource to this resource by the given relation
    /**
     * @template T
     * @param {?} relation
     * @param {?} resource
     * @return {?}
     */
    updateRelation(relation, resource) {
        if (this.existRelationLink(relation)) {
            /** @type {?} */
            const header = ResourceHelper.headers.append('Content-Type', 'text/uri-list');
            return ResourceHelper.getHttp().patch(ResourceHelper.getProxy(this.getRelationLinkHref(relation)), resource._links.self.href, { headers: header });
        }
        else {
            return throwError('no relation found');
        }
    }
    // Bind the given resource to this resource by the given relation
    /**
     * @template T
     * @param {?} relation
     * @param {?} resource
     * @return {?}
     */
    substituteRelation(relation, resource) {
        if (this.existRelationLink(relation)) {
            /** @type {?} */
            const header = ResourceHelper.headers.append('Content-Type', 'text/uri-list');
            return ResourceHelper.getHttp().put(ResourceHelper.getProxy(this.getRelationLinkHref(relation)), resource._links.self.href, { headers: header });
        }
        else {
            return throwError('no relation found');
        }
    }
    // Unbind the resource with the given relation from this resource
    /**
     * @template T
     * @param {?} relation
     * @param {?} resource
     * @return {?}
     */
    deleteRelation(relation, resource) {
        if (this.existRelationLink(relation)) {
            /** @type {?} */
            const link = resource._links.self.href;
            /** @type {?} */
            const idx = link.lastIndexOf('/') + 1;
            if (idx === -1) {
                return throwError('no relation found');
            }
            /** @type {?} */
            const relationId = link.substring(idx);
            return ResourceHelper.getHttp().delete(ResourceHelper.getProxy(this.getRelationLinkHref(relation) + '/' + relationId), { headers: ResourceHelper.headers });
        }
        else {
            return throwError('no relation found');
        }
    }
    /**
     * @private
     * @param {?=} resource
     * @return {?}
     */
    getResourceUrl(resource) {
        /** @type {?} */
        let url = ResourceHelper.getURL();
        if (!url.endsWith('/')) {
            url = url.concat('/');
        }
        if (resource) {
            return url.concat(resource);
        }
        url = url.replace('{?projection}', '');
        return url;
    }
    /**
     * @private
     * @param {?} relation
     * @return {?}
     */
    getRelationLinkHref(relation) {
        if (this._links[relation].templated) {
            return this._links[relation].href.replace('{?projection}', '');
        }
        return this._links[relation].href;
    }
    /**
     * @private
     * @param {?} relation
     * @return {?}
     */
    existRelationLink(relation) {
        return !Utils.isNullOrUndefined(this._links) && !Utils.isNullOrUndefined(this._links[relation]);
    }
}
Resource.decorators = [
    { type: Injectable }
];
/** @nocollapse */
Resource.ctorParameters = () => [];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class AngularHalModule {
    /**
     * @return {?}
     */
    static forRoot() {
        return {
            ngModule: AngularHalModule,
            providers: [
                ExternalService,
                HttpClient,
                {
                    provide: ResourceService,
                    useClass: ResourceService,
                    deps: [ExternalService]
                }
            ]
        };
    }
}
AngularHalModule.decorators = [
    { type: NgModule, args: [{
                imports: [HttpClientModule],
                declarations: [],
                exports: [HttpClientModule],
                providers: [
                    ExternalService,
                    HttpClient,
                    {
                        provide: ResourceService,
                        useClass: ResourceService,
                        deps: [ExternalService]
                    }
                ]
            },] }
];

export { AngularHalModule, CacheHelper, EvictStrategy, ExternalService, Resource, ResourceArray, ResourceHelper, RestService, ResourceService as ɵa };
//# sourceMappingURL=angular4-hal-redux.js.map
