import { Observable } from 'rxjs';
import { Resource } from './resource';
import { ResourceArray } from './resourceArray';
import { Sort } from './sort';
import { SubtypeBuilder } from './subtypeBuilder';
import { Injector } from '@angular/core';
export interface HalParam {
    key: string;
    value: string | number | boolean;
}
export interface HalOptions {
    notPaged?: boolean;
    size?: number;
    sort?: Sort[];
    params?: HalParam[];
}
export declare class RestService<T extends Resource> {
    private injector;
    resourceArray: ResourceArray<T>;
    private type;
    private resource;
    private resourceService;
    private _embedded;
    constructor(type: new () => T, resource: string, injector: Injector, _embedded?: string);
    protected static handleError(error: any): Observable<never>;
    getAll(options?: HalOptions, subType?: SubtypeBuilder): Observable<T[]>;
    get(id: any, params?: HalParam[], builder?: SubtypeBuilder): Observable<T>;
    getBySelfLink(selfLink: string): Observable<T>;
    search(query: string, options?: HalOptions, subType?: SubtypeBuilder): Observable<T[]>;
    searchSingle(query: string, options?: HalOptions): Observable<T>;
    customQuery(query: string, options?: HalOptions, subType?: SubtypeBuilder): Observable<T[]>;
    customQueryPost(query: string, options?: HalOptions, body?: any, subType?: SubtypeBuilder): Observable<T[]>;
    getByRelationArray(relation: string, builder?: SubtypeBuilder): Observable<T[]>;
    getByRelation(relation: string): Observable<T>;
    count(query?: string, options?: HalOptions): Observable<number>;
    create(entity: T): Observable<Observable<never> | T>;
    update(entity: T): Observable<Observable<never> | T>;
    patch(entity: T): Observable<Observable<never> | T>;
    delete(entity: T): Observable<object>;
    totalElement(): number;
    totalPages(): number;
    hasFirst(): boolean;
    hasNext(): boolean;
    hasPrev(): boolean;
    hasLast(): boolean;
    next(): Observable<T[]>;
    prev(): Observable<T[]>;
    first(): Observable<T[]>;
    last(): Observable<T[]>;
    page(pageNumber: number): Observable<T[]>;
    protected handleError(error: any): Observable<never>;
}
